=begin

数式, テキストファイルから
おまかせで GPhys オブジェクトを作成するライブラリです.

= 履歴

* 2012/02/18 Satoshi NODA  ひと通り作成完了
* 2012/02/07 Satoshi NODA  数式から GPhys への変換にオプションを追加
                           テキストデータから GPhys への変換を実装
* 2012/02/06 Satoshi NODA  数式から GPhys への変換を実装

=end

#require 'numru/ggraph'
require "numru/gphys/gphys"
require 'optparse'
include NumRu
include NMath


module NumRu

  # 数式処理. 真面目にやるなら EqnArray クラスを作ると範囲変更などが楽
#  class String
    # 数式の次元を返す
    def eqn_dim
      if self.include_var?('x') then
        ndim = 1
        if self.include_var?('y') then
          ndim = 2
          if self.include_var?('z') then
            ndim = 3
            if self.include_var?('w') then
              ndim = 4
            end
          end
        end
      else
        ndim = 0
        puts "WARNING: #{str} is regarded as constant."
      end
      return ndim
    end


    # 数式 str に変数 var が含まれるかどうか
    def include_var?(var)
      s = self.strip
      if /\W#{var}\W/ =~ s or \
        /^#{var}\W/ =~ s or \
        /\W#{var}$/ =~ s or \
        /^#{var}$/ =~s then
        return true
      else
        return false
      end
    end

    # x,y,z,w 軸の座標を決める
    def gen_xs(ranges=nil, nxs=nil, ndim=nil)
      ndim = 4 if ndim == nil
      if ranges == nil then
        r = [-10.0,10.0]   # default
        puts "MESSAGE: ranges are set default values ([#{r.join(",")}])"
        ranges = Array.new(ndim).fill(r)
      end
      if nxs == nil then
        n = 40             # default
        puts "MESSAGE: number of grid points are set default values (#{n})"
        nxs = Array.new(ndim).fill(n)
      end
      
      xs = Array.new(ndim)
      (0..xs.size-1).each{|i|
        x0 = ranges[i][0]
        dx = (ranges[i][1] - x0).to_f / (nxs[i]-1)
        xs[i] = x0 + dx * NArray.float(nxs[i]).indgen
      }
      xs
    end

    ########################################
    # equation (String) -> NArray
    ########################################
    def eqn2na(str, xs=nil, ranges=nil, nxs=nil)
=begin
         * gphys = eqn(str, 'ranges'=>[[x0,x1],[y0,y1]], 'nxs'=>[nx,ny],
                       'axis_opt'=>[{'long_name'=>'X-axis', 'units'=>'1', 'name'=>'x'},
                           {'long_name'=>'Y-axis', 'units'=>'1', 'name'=>'y'}],
              'data_opt'=>{'long_name'=>'data', 'units'=>'1', 'name'=>'data'}
                 )
    * gphys = eqn(str, 'axis'=>[x_nary, y_nary],
              'axis_opt'=>[{'long_name'=>'X-axis', 'units'=>'1', 'name'=>'x'},
                           {'long_name'=>'Y-axis', 'units'=>'1', 'name'=>'y'}],
              'data_opt'=>{'long_name'=>'data', 'units'=>'1', 'name'=>'data'}
                 )
=end
      
      ndim = str.eqn_dim
      #  puts "MESSAGE: ndim = #{ndim}"

      if xs == nil then
        xs = gen_xs(ranges, nxs, ndim)
      end
            
      # variable name x,y,z,w is important to evaluate equation
      x=xs[0]
      y=xs[1].reshape(1,xs[1].size) if ndim >= 2
      z=xs[2].reshape(1,1,xs[2].size) if ndim >= 3
      w=xs[3].reshape(1,1,1,xs[3].size) if ndim >= 4
      
      data = eval str
      if ndim == 0 then
        puts "WARNING: ndim = 0"
        data = NArray.float(xs.size).fill(data)
      end
      
      return data
    end


    ########################################
    # equation (String) -> GPhys
    ########################################
    def eqn2gphys(str, xs=nil, ranges=nil, nxs=nil, axis_opts=nil, data_opt=nil)

=begin
    * gphys = eqn(str, 'range'=>[[x0,x1],[y0,y1]], 'nx'=>[nx,ny],
              'axis_opt'=>[{'long_name'=>'X-axis', 'units'=>'1', 'name'=>'x'},
                           {'long_name'=>'Y-axis', 'units'=>'1', 'name'=>'y'}],
              'data_opt'=>{'long_name'=>'data', 'units'=>'1', 'name'=>'data'}
                 )
    * gphys = eqn(str, 'axis'=>[x_nary, y_nary],
              'axis_opt'=>[{'long_name'=>'X-axis', 'units'=>'1', 'name'=>'x'},
                           {'long_name'=>'Y-axis', 'units'=>'1', 'name'=>'y'}],
              'data_opt'=>{'long_name'=>'data', 'units'=>'1', 'name'=>'data'}
                 )
=end

      if xs == nil then
        xs = gen_xs(ranges, nxs)
      end
      data = eqn2na(str, xs, ranges, nxs)
      gphys = na2gphys(data, xs, axis_opts, data_opt)
    end
#  end


#  class NArray
    def na2va(nary, opt)
      va = VArray.new( nary,
                       {"long_name"=>opt['long_name'], "units"=>opt['units']},
                       opt['name'] )
    end

    def na2axis(nary, opt)
      va = na2va(nary, opt)
      axis = Axis.new.set_pos(va)
    end

    def axis_default_opts
      opts = Array.new(4)
      opts[0] = {'long_name' => 'X-axis', 'units' => '1', 'name' => 'x'}
      opts[1] = {'long_name' => 'Y-axis', 'units' => '1', 'name' => 'y'}
      opts[2] = {'long_name' => 'Z-axis', 'units' => '1', 'name' => 'z'}
      opts[3] = {'long_name' => 'W-axis', 'units' => '1', 'name' => 'w'}
      opts
    end
    
    ########################################
    # NArray -> GPhys
    ########################################
    def na2gphys(nary, xs, axis_opts=nil, data_opt=nil)
      
      if axis_opts == nil then
        axis_opts = axis_default_opts
      end
      if data_opt == nil
        data_opt = {
          'long_name' => 'data',
          'units' => '1',
          'name' => 'data'
        }
      end
      
      ndim = nary.rank
      axes = Array.new(ndim-1)
      (0..ndim-1).each{|i|
        axes[i] = na2axis(xs[i], axis_opts[i])
      }
      data_a = na2va(nary, data_opt)
      
      # reshape is needed if y is not used
      if ndim == 0 or ndim == 1 then
        gphys = GPhys.new( Grid.new(axes[0]), data_a )
      elsif ndim == 2 then
        gphys = GPhys.new( Grid.new(axes[0],axes[1]), data_a )
      elsif ndim == 3 then
        gphys = GPhys.new( Grid.new(axes[0],axes[1],axes[2]), data_a )
      elsif ndim == 4 then
        gphys = GPhys.new( Grid.new(axes[0],axes[1],axes[2],axes[3]), data_a )
      else
        raise "ERROR: invalid number of dimention #{ndim}"
      end
      
      return gphys
    end
#  end    
    

#  class GPhys
    # gphys を path へ netCDF ファイルとして出力
    def gphys2netcdf(gphys, path)
      outfile = NetCDF.create(path)
      GPhys::IO.write(outfile, gphys)
      outfile.close
      puts "MESSAGE: #{path} is output."
    end
#  end
    
    ########################################
    # path (text file) -> NArray
    ########################################
    def txt2na(path)
      unless File.exist?(path) then
        raise "ERROR: #{path} does not exist."
      end
      
      f = File.open(path)
      
      # gnuplot の入力データ
      #  ary = f.read.split("\n\n").map{ |l| l.split("\n").map{ |ll| ll.split(" ") } }
      tmp = f.read.gsub(/\n{2,}/,"\n")
      ary = tmp.split("\n").map{ |line| line.split(" ") }
      ary = ary.map{|line| line.map{|a| a.to_f}}
      na = NArray.to_na(ary)
    end
    
    
    def extract_col(na, i)
      if i == 0 then
        nary = NArray.new(na.size).indgen
      elsif i > 0 and i <= na.shape[0] then
        nary = na[(i-1),true]
      else
        raise "ERROR: #{i} is invalid number of column."
      end
      nary
    end
    

    ########################################
    # path (text file) -> GPhys
    ########################################
    # 現在は一次元データにのみ対応
    # 将来的には icol_* に配列を入れるようにしたい
    # icol はカラム番号.
    def txt2gphys(path, icol_axis=nil, icol_data=nil, axis_opts=nil, data_opt=nil)
      na = txt2na(path)
      
      ncol = na.shape[0]
      if ncol == 1 then
        icol_axis = 0
        icol_data = 1
      else
        icol_axis = 1
        icol_data = 2
      end
      
      x = extract_col(na, icol_axis)
      data = extract_col(na, icol_data)
      xs = [x]
      
      gphys = na2gphys(data, xs, axis_opts, data_opt)
    end
    
    ########################################
    # Utilities
    ########################################
    # quick open
    # 引数から自動的に内容を判別してファイル読み込みまたは数式処理をして
    # GPhys オブジェクトを返す
    # ファイルは path[@var,[var2]] の形で指定する.
    def qopen(str)
      if str =~ /@/ then
        s, var = str.split('@')
        if var =~ /,/ then
          var = var.split(',')
        end
      else
        s = str
        var = nil
      end
      
      if File.exist?(s) then
        ext = File.extname(s)
        case ext
        when '.nc', '.ctl', '.grib'
          if var == nil then
            var = File.basename(s, ext)
          end
          p "binary (nc, ctl, grib)"
          gphys = GPhys::IO.open(s, var)
        else
          case var.class.to_s
          when NilClass then
            n = 0
          when String then
            n = 1
          when Array then
            n = var.size
          else
            raise "ERROR: #{var} (class: #{var.class}) is invalid."
          end
          
          icol_axis = icol_data = nil
          if n == 1 then
            icol_data = var
          elsif n >= 2 then
            icol_axis = var[0]
            icol_data = var[1]
            if n > 2 then
              puts "WARNING: Too many variables in #{str}. First two is used."
            end
          end
          p "text file"
          gphys = txt2gphys(s, 'icol_axis'=>icol_axis, 'icol_data'=>icol_data)
        end
      else
        # 数式かどうか調べる
        ndim = str.eqn_dim
        if ndim == 0 then
          raise "ERROR: Neither an actual file nor a equation."
        else
          p "equation"
          gphys = eqn2gphys(s)
        end
      end
      gphys
    end

    # おまかせプロット. 1 次元データなら折れ線, 2 次元以上ならトーンとコンター
    def autoplot(gphys, wsn=1)
      if wsn.to_i.abs >= 1 and wsn.to_i.abs <= 4 then
        # drawing (for check)
        DCL.gropn(wsn)

        DCL.sgpset('lcntl', false)
        if gphys.val.rank == 1 then  # ndim = 0 is error
          GGraph.line(gphys)
        else
          GGraph.tone_and_contour(gphys)
          GGraph.color_bar
        end
  
        DCL.grcls
      else
        puts "WARNING: invalid workstation number."
      end
    end

    def autooutput(gphys, out="gphys.nc")
      ext = File.extname(out)
      if ext != '.nc'
        path = out.to_s + '.nc'   # 拡張子自動補完
      else
        path = out.to_s
      end
      gphys2netcdf(gphys, path)  
    end
    
    # quick view
    def qv(str, wsn=1)
      gphys = qopen(str)
      autoplot(gphys, wsn)
    end

    # quick output
    def qout(str, path="gphys.nc")
      gphys = qopen(str)
      autooutput(gphys, path)
    end

end
